import datetime
from typing import List, Union

NUM = Union[int, float]
NUM_LIST = List[NUM]

MAX_SECONDS = 3.154e+9


def get_scaled_time_duration(seconds: float, short: bool = True) -> str:
    """
    Get str formatted duration from number of seconds.milliseconds.
    Output can be in short ar long format, depending on short flag param.
    """
    if seconds < 0:
        return ""
    if seconds >= MAX_SECONDS:
        return "> 100 Years"

    if short:
        sec, msec, min, day, hour, month, year = "s", "ms", "m", "D", "h", "M", "Y"
    else:
        sec, msec, min, day, hour, month, year = "Sec", "MSec", "Min", "Day", "Hour", "Month", "Year"

    d = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=seconds)

    if seconds < 60:
        milliseconds = int(d.microsecond / 1000)
        if milliseconds > 0:
            duration = "{} {} {} {}".format(d.second, sec, milliseconds, msec)
        else:
            duration = "{} {}".format(d.second, sec)
    elif 60 <= seconds < 3600:
        duration = "{} {} {} {}".format(d.minute, min, d.second, sec)
    elif 3600 <= seconds < 24 * 3600:
        duration = "{} {} {} {}".format(d.hour, hour, d.minute, min)
    elif 24 * 3600 <= seconds < 24 * 3600 * 28:
        duration = "{} {} {} {} {} {}".format(d.day - 1, day, d.hour, hour, d.minute, min)
    elif 24 * 3600 * 28 <= seconds < 24 * 3600 * 28 * 12:
        duration = "{} {} {} {} {} {}".format(d.month - 1, month, d.day, day, d.hour, hour)
    else:
        duration = "{} {} {} {} {} {}".format(d.year - 1, year, d.month, month, d.day, day)

    return duration



from copy import copy
from typing import Optional, List, Tuple, Dict


class LiveAxisRange:

    def __init__(self, roll_on_tick: int = 1, offset_left: float = 0., offset_right: float = 0., offset_top: float = 0.,
                 offset_bottom: float = 0., fixed_range: Optional[List[float]] = None) -> None:
        self.roll_on_tick = roll_on_tick
        self.offset_left = offset_left
        self.offset_right = offset_right
        self.offset_top = offset_top
        self.offset_bottom = offset_bottom
        self.crop_left_offset_to_data = False
        self.crop_right_offset_to_data = False
        self.crop_top_offset_to_data = False
        self.crop_bottom_offset_to_data = False
        self.fixed_range = fixed_range
        self.x_range: Dict[str, List[float]] = {}
        self.y_range: Dict[str, List[float]] = {}
        self.final_x_range = [0., 0.]
        self.final_y_range = [0., 0.]
        self.ignored_data_connectors: List[str] = []

    def get_x_range(self, data_connector, tick: int) -> List[float]:
        axis_range = data_connector.plot.data_bounds(ax=0, offset=self.roll_on_tick if self.roll_on_tick > 1 else 0)
        x, _ = data_connector.plot.getData()
        if x is None:
            return [0.]

        final_range = self._get_range(axis_range, tick, (self.offset_left, self.offset_right))
        if final_range is None:
            return self.final_x_range
        offset_x = data_connector.plot.pos().x()
        final_range[0] += offset_x
        final_range[1] += offset_x
        # Check left and right offset and crop to data if flag is set
        if self.crop_left_offset_to_data and final_range[0] < x[0]:
            final_range[0] = x[0]
        if self.crop_right_offset_to_data and final_range[1] > x[-1]:
            final_range[1] = x[-1]
        self.x_range[data_connector.__hash__()] = copy(final_range)
        for connector_id, x_range in self.x_range.items():
            if connector_id in self.ignored_data_connectors:
                continue
            if final_range[0] > x_range[0]:
                final_range[0] = x_range[0]
            if final_range[1] < x_range[1]:
                final_range[1] = x_range[1]
        if final_range[0] == final_range[1]:
            # Pyqtgraph ViewBox.setRange doesn't like same value for min and max,
            # therefore in that case we must set some range
            final_range[0] -= 0.4
            final_range[1] += 0.4
        if self.final_x_range != final_range:
            self.final_x_range = final_range
        return self.final_x_range

    def recalculate_x_range(self):
        final_range = None
        for connector_id, x_range in self.x_range.items():
            if connector_id in self.ignored_data_connectors:
                continue
            elif final_range is None:
                final_range = copy(x_range)
                continue
            if final_range[0] > x_range[0]:
                final_range[0] = x_range[0]
            if final_range[1] < x_range[1]:
                final_range[1] = x_range[1]
        if final_range is None:
            final_range = [0, 0]
        if final_range[0] == final_range[1]:
            # Pyqtgraph ViewBox.setRange doesn't like same value for min and max,
            # therefore in that case we must set some range
            final_range[0] -= 0.4
            final_range[1] += 0.4
        if self.final_x_range != final_range:
            self.final_x_range = final_range
        return self.final_x_range

    def get_y_range(self, data_connector, tick: int) -> List[float]:
        axis_range = data_connector.plot.data_bounds(ax=1, offset=self.roll_on_tick if self.roll_on_tick > 1 else 0)
        _, y = data_connector.plot.getData()
        if y is None:
            return [0.]

        final_range = self._get_range(axis_range, tick, (self.offset_bottom, self.offset_top))
        if final_range is None:
            return self.final_y_range
        offset_y = data_connector.plot.pos().y()
        final_range[0] += offset_y
        final_range[1] += offset_y
        # Check left and right offset and crop to data if flag is set
        if self.crop_bottom_offset_to_data and final_range[0] < y[0]:
            final_range[0] = y[0]
        if self.crop_top_offset_to_data and final_range[1] > y[-1]:
            final_range[1] = y[-1]
        self.y_range[data_connector.__hash__()] = copy(final_range)
        for connector_id, y_range in self.y_range.items():
            if connector_id in self.ignored_data_connectors:
                continue
            if final_range[0] > y_range[0]:
                final_range[0] = y_range[0]
            if final_range[1] < y_range[1]:
                final_range[1] = y_range[1]
        if final_range[0] == final_range[1]:
            # Pyqtgraph ViewBox.setRange doesn't like same value for min and max,
            # therefore in that case we must set some range
            final_range[0] -= 0.4
            final_range[1] += 0.4
        if self.final_y_range != final_range:
            self.final_y_range = final_range
        return self.final_y_range

    def recalculate_y_range(self):
        final_range = None
        for connector_id, y_range in self.y_range.items():
            if connector_id in self.ignored_data_connectors:
                continue
            elif final_range is None:
                final_range = copy(y_range)
                continue
            if final_range[0] > y_range[0]:
                final_range[0] = y_range[0]
            if final_range[1] < y_range[1]:
                final_range[1] = y_range[1]
        if final_range is None:
            final_range = [0, 0]
        if final_range[0] == final_range[1]:
            # Pyqtgraph ViewBox.setRange doesn't like same value for min and max,
            # therefore in that case we must set some range
            final_range[0] -= 0.4
            final_range[1] += 0.4
        if self.final_y_range != final_range:
            self.final_y_range = final_range
        return self.final_y_range

    def _get_range(self, axis_range: Tuple[float, float], tick: int, offsets: Tuple[float, float]) -> Optional[
        List[float]]:
        if self.fixed_range is not None:
            # Return fixed defined range
            return self.fixed_range
        elif self.roll_on_tick == 1:
            # Rolling on every tick
            if offsets[0] + offsets[1] == 0:
                # Return full range if there are no offsets
                return [axis_range[0], axis_range[1]]
            elif tick > 0:
                # Return range of width specified by offsets
                range_width = (abs(axis_range[1] - axis_range[0])) / tick
                return [axis_range[1] - range_width * offsets[0], (axis_range[1] + range_width) + (
                        range_width * offsets[1])]
            else:
                # Just return axis ranges subtracted by offsets
                return [axis_range[0] - offsets[0], axis_range[1] + offsets[1]]
        elif tick % self.roll_on_tick == 0 or tick < 2:
            # Rolling when tick % self.roll_on_tick == 0
            range_width = abs(axis_range[1] - axis_range[0])
            if tick < 2:
                range_width = range_width * (self.roll_on_tick - (tick + 1))
                return [axis_range[1], axis_range[1] + range_width]
            else:
                return [axis_range[1] - range_width * offsets[0], (axis_range[1] + range_width) + (
                        range_width * offsets[1])]
        else:
            return None

    def ignore_connector(self, data_connector, flag: bool) -> None:
        if not flag:
            self.ignored_data_connectors.append(data_connector.__hash__())
        else:
            self.ignored_data_connectors.remove(data_connector.__hash__())
        try:
            self.get_x_range(data_connector, data_connector.rolling_index)
        except TypeError:
            return None
        try:
            self.get_y_range(data_connector, data_connector.rolling_index)
        except TypeError:
            return None

    def remove_data_connector(self, data_connector):
        if data_connector.__hash__() in self.ignored_data_connectors:
            del self.ignored_data_connectors[data_connector.__hash__()]
        if data_connector.__hash__() in self.y_range:
            del self.y_range[data_connector.__hash__()]
        if data_connector.__hash__() in self.x_range:
            del self.x_range[data_connector.__hash__()]



from __future__ import annotations

from typing import Union, Optional, Protocol, Dict, Any, TYPE_CHECKING

import pyqtgraph as pg  # type: ignore
from pyqtgraph.Qt import QtGui, QtCore  # type: ignore

from pglive.kwargs import LeadingLine
from pglive.sources.live_plot_widget import LivePlotWidget
from pglive.sources.utils import NUM_LIST

if TYPE_CHECKING:
    class SupportsLivePlot(Protocol):
        plot_widget: Optional[LivePlotWidget]
        sigPlotChanged: QtCore.Signal
        opts: Dict

        def setData(self, x_data: Any, y_data: Any, **kwargs: Dict) -> None: ...

        def getViewBox(self) -> pg.ViewBox: ...

        def setVisible(self, flag: bool) -> None: ...

else:
    class SupportsLivePlot:
        ...


class MixinLivePlot(SupportsLivePlot):
    """Implements new_data slot for any plot"""
    plot_widget: Optional[LivePlotWidget] = None
    min_x, min_y, max_x, max_y = 0, 0, 0, 0

    def slot_new_data(self, y: NUM_LIST, x: NUM_LIST, kwargs: Dict) -> None:
        self.setData(x, y, **kwargs)

    def slot_connector_toggle(self, data_connector, flag: bool):
        if self.plot_widget is not None:
            self.plot_widget.slot_connector_toggle(data_connector, flag)
        else:
            raise Exception("Plot must be added into LivePlotWidget before setting any data.")

    def slot_roll_tick(self, data_connector, tick: int) -> None:
        if self.plot_widget is not None:
            self.plot_widget.slot_roll_tick(data_connector, tick)
        else:
            raise Exception("Plot must be added into LivePlotWidget before setting any data.")


class MixinLiveBarPlot(SupportsLivePlot):
    """Implements new_data slot for Bar Plot"""
    plot_widget: Optional[LivePlotWidget] = None
    sigPlotChanged = QtCore.Signal()

    def slot_new_data(self, y: Any, x: Any, kwargs: Dict) -> None:
        self.setData(x, y, **kwargs)

    def slot_connector_toggle(self, data_connector, flag: bool):
        if self.plot_widget is not None:
            self.plot_widget.slot_connector_toggle(data_connector, flag)
        else:
            raise Exception("Plot must be added into LivePlotWidget before setting any data.")

    def slot_roll_tick(self, data_connector, tick: int) -> None:
        if self.plot_widget is not None:
            self.plot_widget.slot_roll_tick(data_connector, tick)
        else:
            raise Exception("Plot must be added into LivePlotWidget before setting any data.")


class MixinLeadingLine(SupportsLivePlot):
    """Implements leading line"""
    _hl_kwargs = None
    _vl_kwargs = None

    def set_leading_line(self, orientation: str = LeadingLine.VERTICAL,
                         pen: QtGui.QPen = None, text_axis: str = LeadingLine.AXIS_X, **kwargs) -> Dict:
        text_axis = text_axis.lower()
        assert text_axis in (LeadingLine.AXIS_X, LeadingLine.AXIS_Y)

        self.sigPlotChanged.connect(self.update_leading_line)

        if pen is None:
            pen = self.opts.get("pen")
        if orientation == LeadingLine.VERTICAL:
            _v_leading_line = pg.InfiniteLine(angle=90, movable=False, pen=pen)
            _v_leading_text = pg.TextItem(color="black", angle=-90, fill=pen.color())
            _v_leading_line.setZValue(999)
            _v_leading_text.setZValue(999)
            self._vl_kwargs = {"line": _v_leading_line, "text": _v_leading_text, "pen": pen, "text_axis": text_axis,
                               **kwargs}
            return self._vl_kwargs
        elif orientation == LeadingLine.HORIZONTAL:
            _h_leading_line = pg.InfiniteLine(angle=0, movable=False, pen=pen)
            _h_leading_text = pg.TextItem(color="black", fill=pen.color())
            _h_leading_text.setZValue(999)
            _h_leading_text.setZValue(999)
            self._hl_kwargs = {"line": _h_leading_line, "text": _h_leading_text, "pen": pen, "text_axis": text_axis,
                               **kwargs}
            return self._hl_kwargs
        else:
            raise TypeError("Unsupported LeadingLine type")

    def update_leading_line(self):
        raise NotImplementedError

    def x_format(self, value: Union[int, float]) -> str:
        """X tick format (will be overwritten when inserted in LivePlotWidget)"""
        return str(round(value, 4))

    def y_format(self, value: Union[int, float]) -> str:
        """Y tick format (will be overwritten when inserted in LivePlotWidget)"""
        return str(round(value, 4))

    def update_leading_text(self, x: float, y: float, x_text: Optional[str] = None,
                            y_text: Optional[str] = None) -> None:
        """Update position and text of Vertical and Horizontal leading text"""
        vb = self.getViewBox()
        width, height = vb.width(), vb.height()
        if x_text is None:
            x_text = self.x_format(x)
        if y_text is None:
            y_text = self.y_format(y)

        if self._vl_kwargs is not None:
            text_axis = x_text if self._vl_kwargs["text_axis"] == LeadingLine.AXIS_X else y_text
            self._vl_kwargs["text"].setText(text_axis)
            pixel_pos = vb.mapViewToScene(QtCore.QPointF(x, y))
            y_pos = 0 + self._vl_kwargs["text"].boundingRect().height() + 10
            new_pos = vb.mapSceneToView(QtCore.QPointF(pixel_pos.x(), y_pos))
            self._vl_kwargs["text"].setPos(new_pos.x(), new_pos.y())

        if self._hl_kwargs is not None:
            text_axis = x_text if self._hl_kwargs["text_axis"] == LeadingLine.AXIS_X else y_text
            self._hl_kwargs["text"].setText(text_axis)
            pixel_pos = vb.mapViewToScene(QtCore.QPointF(x, y))
            x_pos = width - self._hl_kwargs["text"].boundingRect().width() + 21
            new_pos = vb.mapSceneToView(QtCore.QPointF(x_pos, pixel_pos.y()))
            self._hl_kwargs["text"].setPos(new_pos.x(), new_pos.y())

    def clear_leading_lines(self):
        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(0)
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(0)
        self.update_leading_text(0, 0)



import signal
from threading import Thread

import pglive.examples_pyqt5 as examples
import pyqtgraph as pg   # type: ignore
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_axis_range import LiveAxisRange
from pglive.sources.live_plot import LiveLinePlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Range crop offset to data is displayed In this example.
"""
layout = pg.LayoutWidget()
layout.layout.setSpacing(0)

'''
We want to display 30 seconds long view and pan right every 1 second. 
'''
widget = LivePlotWidget(title="Roll plot view every 1 sec, offset 30 sec, crop left = False",
                        x_range_controller=LiveAxisRange(roll_on_tick=100, offset_left=30),
                        y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
plot = LiveLinePlot(pen="red")
widget.addItem(plot)
layout.addWidget(widget, row=0, col=0)
data_connector = DataConnector(plot, max_points=6000)

'''
Move view to the right on every 300 ticks (data update).
Y range is fixed to -1 and 1.
'''
widget2 = LivePlotWidget(title="Roll plot view every 1 sec, offset 30 sec, crop left = True",
                         x_range_controller=LiveAxisRange(roll_on_tick=100, offset_left=30),
                         y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
widget2.x_range_controller.crop_left_offset_to_data = True
plot2 = LiveLinePlot(pen="red")
widget2.addItem(plot2)
layout.addWidget(widget2, row=1, col=0)
data_connector2 = DataConnector(plot2, max_points=6000)

'''
Move view to the right on every 300 ticks (data update).
Y range is fixed to -1 and 1.
'''
widget3 = LivePlotWidget(title="Roll plot view every 1 sec, offset 30 sec, crop bottom = False",
                         y_range_controller=LiveAxisRange(roll_on_tick=100, offset_bottom=30),
                         x_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
plot3 = LiveLinePlot(pen="yellow")
widget3.addItem(plot3)
layout.addWidget(widget3, row=0, col=1)
data_connector3 = DataConnector(plot3, max_points=6000)

'''
Move view to the right on every 300 ticks (data update).
Y range is fixed to -1 and 1.
'''
widget4 = LivePlotWidget(title="Roll plot view every 1 sec, offset 30 sec, crop bottom = True",
                         y_range_controller=LiveAxisRange(roll_on_tick=100, offset_bottom=30),
                         x_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
widget4.y_range_controller.crop_bottom_offset_to_data = True
plot4 = LiveLinePlot(pen="yellow")
widget4.addItem(plot4)
layout.addWidget(widget4, row=1, col=1)
data_connector4 = DataConnector(plot4, max_points=6000)

layout.show()

Thread(target=examples.sin_wave_generator, args=(data_connector, data_connector2)).start()
Thread(target=examples.sin_wave_generator, args=(data_connector3, data_connector4), kwargs={"flip": True}).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



from typing import Dict, Any, Tuple, List

import numpy as np  # type: ignore
import pyqtgraph as pg  # type: ignore

from pglive.sources.live_mixins import MixinLivePlot, MixinLeadingLine, MixinLiveBarPlot


class LiveLinePlot(pg.PlotDataItem, MixinLivePlot, MixinLeadingLine):
    """Line plot"""

    def update_leading_line(self) -> None:
        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(self.xData[-1])

        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(self.yData[-1])
        self.update_leading_text(self.xData[-1], self.yData[-1])

    def clear(self):
        try:
            self.clear_leading_lines()
        except:
            pass

        super().clear()

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple:
        x, y = self.getData()
        if x is None and y is None:
            return 0, 0
        if ax == 0:
            sub_range = x[-offset:]
        else:
            sub_range = y[-offset:]
        return np.nanmin(sub_range), np.nanmax(sub_range)


class LiveScatterPlot(pg.ScatterPlotItem, MixinLivePlot, MixinLeadingLine):
    """Scatter plot"""

    def clear(self):
        try:
            self.clear_leading_lines()
        except AttributeError:
            pass

        super().clear()

    def update_leading_line(self) -> None:
        last_point = self.data[-1]
        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(last_point[0])
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(last_point[1])

        self.update_leading_text(last_point[0], last_point[1])

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        x, y = self.getData()
        if x.size == 0 and y.size == 0:
            return 0, 0
        if ax == 0:
            sub_range = x[-offset:]
        else:
            sub_range = y[-offset:]
        return np.nanmin(sub_range), np.nanmax(sub_range)


class LiveHBarPlot(pg.BarGraphItem, MixinLiveBarPlot, MixinLeadingLine):
    """Horizontal Bar Plot"""

    def __init__(self, x0: float = 0., bar_height: float = 1., **kwargs: Any) -> None:
        self.bar_height = bar_height
        self.x0 = x0
        super().__init__(x0=x0, y=[0], width=0, height=0, **kwargs)

    def setData(self, x_data: float, y_data: float, **kwargs: Dict) -> None:
        self.setOpts(x0=self.x0, y=x_data, height=self.bar_height, width=y_data, **kwargs)
        self.sigPlotChanged.emit()

    def clear(self):
        self.setOpts(x0=self.x0, y=[], height=self.bar_height, width=[])
        self.sigPlotChanged.emit()

    def getData(self) -> Tuple[List[float], List[float]]:
        return self.opts["width"], self.opts["y"]

    def update_leading_line(self) -> None:
        if self.opts["width"] == [] and self.opts["y"] == []:
            self.clear_leading_lines()
            return

        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(self.opts["width"][-1])
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(self.opts["y"][-1])
        self.update_leading_text(self.opts["width"][-1], self.opts["y"][-1])

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        x, y = self.getData()
        if x == [] and y == []:
            return 0, 0
        if ax == 0:
            sub_range = x[-offset:]
        else:
            sub_range = y[-offset:]
        return np.nanmin(sub_range), np.nanmax(sub_range)


class LiveVBarPlot(pg.BarGraphItem, MixinLiveBarPlot, MixinLeadingLine):
    """Vertical Bar Plot"""

    def __init__(self, y0: float = 0, bar_width: float = 1, **kwargs: Any) -> None:
        self.bar_width = bar_width
        self.y0 = y0
        super().__init__(y0=y0, x=[0], width=0, height=0, **kwargs)

    def setData(self, x_data: float, y_data: float, **kwargs: Dict) -> None:
        self.setOpts(y0=self.y0, x=x_data, height=y_data, width=self.bar_width, **kwargs)
        self.sigPlotChanged.emit()

    def clear(self):
        self.setOpts(y0=self.y0, x=[], height=[], width=self.bar_width)
        self.sigPlotChanged.emit()

    def update_leading_line(self) -> None:
        if self.opts["x"] == [] and self.opts["height"] == []:
            self.clear_leading_lines()
            return

        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(self.opts["x"][-1])
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(self.opts["height"][-1])
        self.update_leading_text(self.opts["x"][-1], self.opts["height"][-1])

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        x, y = self.getData()
        if x == [] and y == []:
            return 0, 0
        if ax == 0:
            sub_range = x[-offset:]
        else:
            sub_range = y[-offset:]
        return np.nanmin(sub_range), np.nanmax(sub_range)


def make_live(plot: pg.GraphicsObject) -> None:
    """Convert plot into Live plot"""
    if isinstance(plot, pg.BarGraphItem):
        # Create horizontal bar plot in case of bar plot
        plot.bar_width = 1
        plot.y0 = 0
        plot.slot_new_data = lambda y, x, kwargs: plot.setOpts(y0=plot.y0, x=x, height=y, width=plot.bar_width,
                                                               **kwargs)
    elif isinstance(plot, pg.PlotCurveItem):
        plot.slot_new_data = lambda y, x, kwargs: plot.setData(np.array(x), np.array(y), **kwargs)
    else:
        plot.slot_new_data = lambda y, x, kwargs: plot.setData(x, y, **kwargs)

    def data_bounds(ax: int = 0, offset: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        x, y = plot.getData()
        if x is None and y is None:
            return 0, 0
        if ax == 0:
            sub_range = x[-offset:]
        else:
            sub_range = y[-offset:]
        return np.nanmin(sub_range), np.nanmax(sub_range)

    plot.data_bounds = data_bounds
    plot.slot_connector_toggle = lambda data_connector, flag: plot.plot_widget.slot_connector_toggle(data_connector,
                                                                                                     flag)
    plot.slot_roll_tick = lambda data_connector, tick: plot.plot_widget.slot_roll_tick(data_connector, tick)



import signal
from threading import Thread

import pglive.examples_pyqt5 as examples
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_candleStickPlot import LiveCandleStickPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Candlestick Bar plot is displayed in this example.
"""
win = LivePlotWidget(title="Candlestick Plot @ 10Hz")
plot = LiveCandleStickPlot()
win.addItem(plot)

data_connector = DataConnector(plot, max_points=50, update_rate=10)

win.show()

Thread(target=examples.candle_generator, args=(data_connector,)).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



import pglive.examples_pyqt6 as examples
from pyqtgraph import mkPen  # type: ignore

from threading import Thread

import pyqtgraph as pg  # type: ignore

from pglive.kwargs import LeadingLine
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveLinePlot, LiveScatterPlot, LiveHBarPlot, LiveVBarPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Leading line is displayed in this example.
Pglive can plot Vertical or Horizontal leading line.
User can choose orientation and which value is displayed by using text_axis parameter.
Please note, that X might be swapped with Y, when plotting horizontally or vertically.
Color of leading line is set with pen parameter.
"""

layout = pg.LayoutWidget()
# Line plot displaying Vertical Leading line and using Y axis value
plot_widget_1 = LivePlotWidget(title="Line Plot @ 50Hz, Leading with Y")
plot = LiveLinePlot()
plot.set_leading_line(LeadingLine.VERTICAL, pen=mkPen("red"), text_axis=LeadingLine.AXIS_Y)
plot_widget_1.addItem(plot)
data_connector = DataConnector(plot, max_points=300, update_rate=50)
layout.addWidget(plot_widget_1)

# Line plot displaying Horizontal Leading line and using Y axis value
# Note that in this case text_axis default is X, but we have flipped x and y values.
plot_widget_2 = LivePlotWidget(title="Line Plot @ 50Hz, Leading with Y")
plot = LiveScatterPlot()
plot.set_leading_line(LeadingLine.HORIZONTAL, pen=mkPen("yellow"))
plot_widget_2.addItem(plot)
data_connector2 = DataConnector(plot, max_points=300, update_rate=50)
layout.addWidget(plot_widget_2)

plot_widget_3 = LivePlotWidget(title="Line Plot @ 50Hz, Leading with X")
plot = LiveHBarPlot()
plot.set_leading_line(LeadingLine.HORIZONTAL, pen=mkPen("yellow"), text_axis=LeadingLine.AXIS_Y)
plot_widget_3.addItem(plot)
data_connector3 = DataConnector(plot, max_points=300, update_rate=50)
layout.addWidget(plot_widget_3)

plot_widget_4 = LivePlotWidget(title="Line Plot @ 50Hz, Leading with Y")
plot = LiveVBarPlot()
plot.set_leading_line(LeadingLine.HORIZONTAL, pen=mkPen("red"), text_axis=LeadingLine.AXIS_Y)
plot_widget_4.addItem(plot)
data_connector4 = DataConnector(plot, max_points=300, update_rate=50)
layout.addWidget(plot_widget_4)

layout.show()

Thread(target=examples.sin_wave_generator, args=(data_connector, data_connector3, data_connector4)).start()
Thread(target=examples.sin_wave_generator, args=(data_connector2, ), kwargs={"flip": True}).start()
examples.app.exec()
examples.stop()



from threading import Thread

import pglive.examples_pyqt5 as examples
import pyqtgraph as pg  # type: ignore

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveHBarPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Different plot rate is demonstrated in this example.
Display four plots, each slower by 1/4 from previous plot rate.
Plot rate is set in Hz unit.
"""
layout = pg.LayoutWidget()
args = []
# Initial rate of 100Hz
plot_rate = 100.
max_len = 600
# Initial delta Y is 1
bar_height = 1
for index in range(4):
    widget = LivePlotWidget(title=f"Horizontal Bar Plot @ {plot_rate}Hz")
    plot = LiveHBarPlot(bar_height=bar_height, brush="green", pen="green")
    widget.addItem(plot)
    layout.addWidget(widget)
    args.append(DataConnector(plot, max_points=max_len, plot_rate=plot_rate))
    plot_rate /= 4.

layout.show()

Thread(target=examples.sin_wave_generator, args=args).start()
examples.app.exec()
examples.stop()



import copy
import time
import warnings
from collections import deque
from math import inf
from threading import Lock
from typing import List, Union, Deque, Optional

import numpy as np
from pyqtgraph import PlotDataItem  # type: ignore
from pyqtgraph.Qt import QtCore  # type: ignore

from pglive.sources.live_plot import MixinLivePlot, MixinLiveBarPlot, make_live
from pglive.sources.utils import NUM_LIST, NUM

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)


class DataConnector(QtCore.QObject):
    sig_new_data = QtCore.Signal(object, object, dict)
    sig_data_roll_tick = QtCore.Signal(object, int)
    sig_data_toggle = QtCore.Signal(object, bool)
    sig_paused = QtCore.Signal()
    sig_resumed = QtCore.Signal()
    sig_clear = QtCore.Signal()
    paused: bool = False
    # Last update time, using perf_counter for most precise counter
    last_update: float = 0.
    last_plot: float = 0.

    def __init__(self, plot: Union[MixinLivePlot, MixinLiveBarPlot], max_points: float = inf, update_rate: float = inf,
                 plot_rate: float = inf, ignore_auto_range: bool = False) -> None:
        """
        DataConnector is connecting plot with data and makes sure, that all updates are thread-safe.
        To make plot compatible and work with Connector, it must implement slot_new_data method.
        This can be achieved either by inheriting from LiveMixin or LiveMixinBarPlot class or DataConnector, tries to
        use make_live function and add slot in runtime.
        Use ignore_auto_range if you have more than one plot to calculate range for only one main plot.
        This can also improve performance for multiple plots inside one widget.
        :param plot: Plot to be connected with Data
        :param max_points: Maximum amount of data points to plot
        :param float update_rate: Update rate in Hz
        :param float plot_rate: Plot rate in Hz
        :param bool ignore_auto_range: If set to True auto range is not calculated when new data is acquired
        """
        super().__init__()
        self.rolling_index = 0
        self.ignore_auto_range = ignore_auto_range

        if not isinstance(plot, (MixinLivePlot, MixinLiveBarPlot)):
            # Attempt to convert plot into live if it's not already
            make_live(plot)
        # Data update lock, ensuring thread-safety
        self.data_lock = Lock()
        # Maximum number of points to plot
        self.max_points = max_points
        # Calculating update timeout from update_rate frequency
        self.update_timeout = 1 / update_rate
        self.plot_timeout = 1 / plot_rate
        self.tick_position_indexes = None
        self.plot = plot
        # Set plot and connect sig_new_data with plot.slot_new_data
        self.sig_new_data.connect(self.plot.slot_new_data)
        self.sig_data_toggle.connect(self.plot.slot_connector_toggle)
        self.sig_data_roll_tick.connect(self.plot.slot_roll_tick)
        self.sig_clear.connect(self.plot.clear)
        self.x: Union[NUM_LIST, Deque[NUM], List]
        self.y: Union[NUM_LIST, Deque[NUM], List]
        if self.max_points == inf:
            # Use simple list if there is no point limits
            self.x = []
            self.y = []
        else:
            # Use deque with maxlen otherwise
            self.x = deque(maxlen=int(self.max_points))
            self.y = deque(maxlen=int(self.max_points))

        def toggle_plot_visibility(flag):
            """Override setVisible of PlotDataItem"""
            PlotDataItem.setVisible(self.plot, flag)
            self.sig_data_toggle.emit(self, flag)

        setattr(self.plot, "data_connector", self)
        setattr(self.plot, "setVisible", toggle_plot_visibility)

    @property
    def max_points(self) -> Union[int, float]:
        return self._max_len

    @max_points.setter
    def max_points(self, new_max_len: Union[int, float]) -> None:
        assert new_max_len > 0
        self._max_len = new_max_len

    def pause(self) -> None:
        """Pause data plotting"""
        self.paused = True
        self.sig_paused.emit()

    def resume(self) -> None:
        """Resume data plotting"""
        self.paused = False
        self.sig_resumed.emit()

    def clear(self) -> None:
        """Clear all data"""
        with self.data_lock:
            self.x.clear()
            self.y.clear()
            self.rolling_index = 0
            self.tick_position_indexes = None
            self.sig_clear.emit()
            self.sig_data_roll_tick.emit(self, 0)

    def _skip_update(self) -> bool:
        """Skip data update"""
        return self.paused or (time.perf_counter() - self.last_update) < self.update_timeout or self.data_lock.locked()

    def _skip_plot(self) -> bool:
        """Skip data plot"""
        return self.paused or (time.perf_counter() - self.last_plot) < self.plot_timeout

    def _update_data(self, **kwargs):
        """Update data and last update time"""
        # Notify all connected plots
        try:
            self.sig_new_data.emit(np.asarray(self.y), np.asarray(self.x), kwargs)
        except ValueError:
            self.sig_new_data.emit(copy.copy(self.y), np.asarray(self.x), kwargs)
        self.last_plot = time.perf_counter()

    def cb_set_data(self, y: List[Union[int, float]], x: Optional[NUM_LIST] = None, **kwargs) -> None:
        """Replace current data"""
        if self._skip_update():
            return

        with self.data_lock:
            if self.max_points == inf:
                self.y = y
            else:
                self.y = deque(y, maxlen=int(self.max_points))
            if x is not None:
                if self.max_points == inf:
                    self.x = x
                else:
                    self.x = deque(x, maxlen=int(self.max_points))
            else:
                self.x = list(range(len(self.y)))

            self.last_update = time.perf_counter()

            if not self._skip_plot():
                self._update_data(**kwargs)
                self.sig_data_roll_tick.emit(self, len(self.x) - 1)
                self.rolling_index = len(self.x)

    def cb_append_data_point(self, y: Union[int, float], x: Optional[Union[int, float]] = None, **kwargs) -> None:
        """Append new data point"""
        if self._skip_update():
            return

        with self.data_lock:
            self.y.append(y)
            if x is not None:
                self.x.append(x)
            elif len(self.x) == 0:
                self.x.append(0)
            else:
                self.x.append(self.x[-1] + 1)
            if self.tick_position_indexes is not None:
                if len(self.tick_position_indexes) == 0:
                    self.tick_position_indexes.append(0.0)
                elif len(self.tick_position_indexes) == self.tick_position_indexes.maxlen:
                    self.tick_position_indexes.rotate(-1)
                else:
                    self.tick_position_indexes.append(self.tick_position_indexes[-1] + 1.0)
            self.last_update = time.perf_counter()

            if not self._skip_plot():
                self._update_data(**kwargs)
                self.sig_data_roll_tick.emit(self, self.rolling_index)
                self.rolling_index += 1

    def cb_append_data_array(self, y: List[Union[int, float]], x: Optional[NUM_LIST] = None, **kwargs) -> None:
        """Append array of data to existing dataset"""
        if self._skip_update():
            return

        with self.data_lock:
            if self.max_points == inf:
                # Concat two lists
                self.y += y
                if x is not None:
                    self.x += x
                else:
                    self.x += list(range(self.x[-1], self.x[-1] + len(y), 1))
            else:
                # Concat two queues
                self.y += deque(y)
                if x is not None:
                    self.x += deque(x)
                else:
                    self.x += deque(range(self.x[-1], self.x[-1] + len(y), 1))

            if self.tick_position_indexes is not None:
                if len(self.tick_position_indexes) == 0:
                    self.tick_position_indexes.append(0.0)
                elif len(self.tick_position_indexes) == self.tick_position_indexes.maxlen:
                    self.tick_position_indexes.rotate(-1)
                else:
                    self.tick_position_indexes.append(self.tick_position_indexes[-1] + 1.0)
            self.last_update = time.perf_counter()

            if not self._skip_plot():
                self._update_data(**kwargs)
                self.sig_data_roll_tick.emit(self, self.rolling_index)
                self.rolling_index += len(y)



from typing import List, Tuple, Optional, Dict, Any

import numpy as np
import pyqtgraph as pg  # type: ignore
from pyqtgraph.Qt import QtCore, QtGui  # type: ignore

from pglive.sources.live_mixins import MixinLivePlot, MixinLeadingLine


class LiveCategorizedBarPlot(pg.GraphicsObject, MixinLivePlot, MixinLeadingLine):
    """Live categorized bar plot, plotting data [[category1, category2, ...], ...]"""
    sigPlotChanged = QtCore.Signal(object)

    def __init__(self, categories: Optional[List[str]] = None, category_color: Optional[Dict[str, str]] = None,
                 bar_height: float = 0.9) -> None:
        pg.GraphicsObject.__init__(self)
        self.x_data: List[float] = []
        self.y_data: List[Tuple[str]] = []
        self.bar_height = bar_height
        self.outline_pen = pg.mkPen("w")
        self.bar_brush = pg.mkBrush("g")
        self.categories = categories if categories is not None else []
        self.category_color = {category: pg.mkBrush(color) for category, color in
                               category_color.items()} if category_color is not None else {}
        self.picture = QtGui.QPicture()

    @property
    def bar_height(self) -> float:
        return self._bar_height

    @bar_height.setter
    def bar_height(self, new_bar_height: float) -> None:
        assert 0 <= new_bar_height <= 1
        self._bar_height = new_bar_height

    def paint(self, p: QtGui.QPainter, *args: Any) -> None:
        p.drawPicture(0, 0, self.picture)

    def boundingRect(self) -> QtCore.QRect:
        return QtCore.QRectF(self.picture.boundingRect())

    def clear(self):
        self.x_data = []
        self.y_data = []
        self.output_y_data = []
        self.picture = QtGui.QPicture()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.bounds = [None, None]
        self.sigPlotChanged.emit(self)

    def setData(self, x_data: List[float], y_data: List[Tuple[str]], **kwargs: Dict) -> None:
        """y_data must be in format [[category1, category2, ...], ...]"""
        self.x_data = x_data
        self.y_data = y_data
        if len(x_data) != len(y_data):
            raise Exception("Len of x_data must be the same as y_data")

        self.picture = QtGui.QPicture()
        p = QtGui.QPainter(self.picture)
        p.setPen(self.outline_pen)

        active_categories = {}
        for index, x in enumerate(x_data):
            for category in y_data[index]:
                if category not in self.categories:
                    self.categories.append(category)
                    active_categories[category] = x
                if category not in active_categories:
                    active_categories[category] = x
            for category in tuple(active_categories.keys()):
                if category not in y_data[index]:
                    try:
                        p.setBrush(self.category_color[category])
                    except KeyError:
                        p.setBrush(self.bar_brush)
                    w = x - active_categories[category]
                    p.drawRect(
                        QtCore.QRectF(active_categories[category],
                                      self.categories.index(category) - self.bar_height / 2, w, self.bar_height))
                    del active_categories[category]
        for category in active_categories.keys():
            try:
                p.setBrush(self.category_color[category])
            except KeyError:
                p.setBrush(self.bar_brush)
            w = x - active_categories[category]
            p.drawRect(
                QtCore.QRectF(active_categories[category], self.categories.index(category) - self.bar_height / 2, w,
                              self.bar_height))

        p.end()

        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.bounds = [None, None]
        self.sigPlotChanged.emit(self)

    def update_leading_line(self) -> None:
        """Leading line will display all four fields"""
        last_x_point = self.x_data[-1]
        last_y_point = len(self.categories) - 1
        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(last_x_point)
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(last_y_point)

        y_text = ", ".join(self.y_data[-1])
        self.update_leading_text(last_x_point, last_y_point, y_text=y_text)

    def getData(self):
        return self.x_data, self.y_data

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple:
        if ax == 0:
            if self.x_data == []:
                return 0, 0
            sub_range = self.x_data[-offset:]
            return np.nanmin(sub_range), np.nanmax(sub_range)
        else:
            h = self.bar_height / 2
            return 0 - h, len(self.categories) - (1 - h)



import pglive.examples_pyside6 as examples
import signal
from threading import Thread

from PySide6.QtWidgets import QWidget, QGridLayout, QPushButton, QLabel

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveScatterPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Pause and Resume functionality of DataConnector is demonstrated in this example.
There are two buttons, that pause and resume live plotting.
When Live plot is paused, data are not collected.
"""
# Create parent widget
widget = LivePlotWidget(title="Line Plot @ 100Hz")
plot = LiveScatterPlot()
widget.addItem(plot)

# Create plot widget
parent_widget = QWidget()
parent_layout = QGridLayout()
parent_widget.setLayout(parent_layout)

# Connect plot with DataConnector
data_connector = DataConnector(plot, max_points=600)

# Create Pause, Resume buttons and Live status label
pause_button = QPushButton("Pause live plot")
resume_button = QPushButton("Resume live plot")
status_label = QLabel("Live")

# Add widgets int parent widget
parent_layout.addWidget(widget)
parent_layout.addWidget(pause_button)
parent_layout.addWidget(resume_button)
parent_layout.addWidget(status_label)

# Connect signals with respective methods
pause_button.clicked.connect(data_connector.pause)
resume_button.clicked.connect(data_connector.resume)

# Connect sig_paused and sig_resume with live signal status label
data_connector.sig_paused.connect(lambda: status_label.setText("Paused"))
data_connector.sig_resumed.connect(lambda: status_label.setText("Live"))

parent_widget.show()

Thread(target=examples.sin_wave_generator, args=(data_connector,)).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



from typing import List, Tuple, Any, Dict

import numpy as np
import pyqtgraph as pg  # type: ignore
from numpy import ndarray
from pyqtgraph.Qt import QtCore, QtGui  # type: ignore

from pglive.sources.live_mixins import MixinLivePlot, MixinLeadingLine


class LiveCandleStickPlot(pg.GraphicsObject, MixinLivePlot, MixinLeadingLine):
    """Live candlestick plot, plotting data [[open, close, min, max], ...]"""
    sigPlotChanged = QtCore.Signal(object)

    def __init__(self, outline_color: str = "w", high_color: str = 'g', low_color: str = 'r') -> None:
        """Choose colors of candle"""
        pg.GraphicsObject.__init__(self)
        self.x_data: List[float] = []
        self.y_data: List[Tuple[float, ...]] = []
        self.output_y_data: List[float] = []
        self.outline_pen = pg.mkPen(outline_color)
        self.high_brush = pg.mkBrush(high_color)
        self.low_brush = pg.mkBrush(low_color)
        self.picture = QtGui.QPicture()

    def paint(self, p: QtGui.QPainter, *args: Any) -> None:
        p.drawPicture(0, 0, self.picture)

    def boundingRect(self) -> QtCore.QRect:
        return QtCore.QRectF(self.picture.boundingRect())

    def clear(self):
        self.x_data = []
        self.y_data = []
        self.output_y_data = []
        self.picture = QtGui.QPicture()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.bounds = [None, None]
        self.sigPlotChanged.emit(self)

    def setData(self, x_data: List[float], y_data: List[Tuple[float, ...]], **kwargs: Dict) -> None:
        """y_data must be in format [[open, close, min, max], ...]"""
        self.x_data = x_data
        self.y_data = y_data
        self.output_y_data = []
        if len(x_data) != len(y_data):
            raise Exception("Len of x_data must be the same as y_data")

        self.picture = QtGui.QPicture()
        p = QtGui.QPainter(self.picture)
        p.setPen(self.outline_pen)
        try:
            w = (x_data[1] - x_data[0]) / 3.
        except IndexError:
            w = 1 / 3

        for index, (open, close, min, max) in enumerate(y_data):
            t = x_data[index]
            p.drawLine(QtCore.QPointF(t, min), QtCore.QPointF(t, max))
            if open > close:
                p.setBrush(self.low_brush)
            else:
                p.setBrush(self.high_brush)
            p.drawRect(QtCore.QRectF(t - w, open, w * 2, close - open))
            self.output_y_data.extend([min, max])
        p.end()

        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.bounds = [None, None]
        self.sigPlotChanged.emit(self)

    def update_leading_line(self) -> None:
        """Leading line will display all four fields"""
        last_x_point = self.x_data[-1]
        last_y_point = self.y_data[-1][0]
        if self._vl_kwargs is not None:
            self._vl_kwargs["line"].setPos(last_x_point)
        if self._hl_kwargs is not None:
            self._hl_kwargs["line"].setPos(last_y_point)

        y_text = str([round(x, 4) for x in self.y_data[-1]])
        self.update_leading_text(last_x_point, last_y_point, y_text=y_text)

    def getData(self) -> Tuple[List[float], List[Tuple[float, ...]]]:
        return self.x_data, self.y_data

    def data_bounds(self, ax: int = 0, offset: int = 0) -> Tuple[ndarray, ndarray]:
        if self.x_data == [] and self.output_y_data == []:
            return 0, 0
        if ax == 0:
            sub_range = self.x_data[-offset:]
        else:
            sub_range = self.output_y_data[-offset * 2:]
        return np.nanmin(sub_range), np.nanmax(sub_range)



import random
import signal
import sys
from math import sin, cos
from time import sleep
from typing import List

from PySide6.QtGui import QColor
from PySide6.QtWidgets import QApplication

running = True
app = QApplication(sys.argv)


def stop():
    """Stop current QApplication"""
    global running
    running = False
    app.exit(0)


def sin_wave_generator(*data_connectors, flip=False):
    """Sine wave generator"""
    x = 0
    while running:
        x += 1
        for data_connector in data_connectors:
            if flip:
                data_connector.cb_append_data_point(x, sin(x * 0.025))
            else:
                data_connector.cb_append_data_point(sin(x * 0.025), x)
        sleep(0.01)


def sin_wave_array_generator(*data_connectors, flip=False, points=10):
    """Sine wave generator"""
    x = 0
    while running:
        for data_connector in data_connectors:
            if flip:
                data_connector.cb_append_data_array(list(range(x, x + points)),
                                                    [sin((x + xi) * 0.025) for xi in range(points)])
            else:
                data_connector.cb_append_data_array([sin((x + xi) * 0.025) for xi in range(points)],
                                                    list(range(x, x + points)))
        x += points
        sleep(0.01)


def cos_wave_generator(*data_connectors, flip=False):
    """Cosine wave generator"""
    x = 0
    while running:
        x += 1
        for data_connector in data_connectors:
            if flip:
                data_connector.cb_append_data_point(x, cos(x * 0.025))
            else:
                data_connector.cb_append_data_point(cos(x * 0.025), x)
        sleep(0.01)


def candle_generator(*data_connectors, flip=False):
    """Candle stick generator"""
    x = 0
    while running:
        x += 1
        for data_connector in data_connectors:
            a, b = sin(x * 0.025), sin(x * 0.020)
            s = min(a, b) - random.randint(0, 1000) * 1e-3
            e = max(a, b) + random.randint(0, 1000) * 1e-3
            candle = (a, b, s, e)
            if flip:
                data_connector.cb_append_data_point(x, candle)
            else:
                data_connector.cb_append_data_point(candle, x)
        sleep(0.01)


def candle_array_generator(*data_connectors, flip=False, points=10):
    """Candle stick generator"""
    x = 0

    def candle(xc):
        a, b = sin(xc * 0.025), sin(xc * 0.020)
        s = min(a, b) - random.randint(0, 1000) * 1e-3
        e = max(a, b) + random.randint(0, 1000) * 1e-3
        return (a, b, s, e)

    while running:
        for data_connector in data_connectors:
            if flip:
                data_connector.cb_append_data_array(list(range(x, x + points)),
                                                    [candle(x + xi) for xi in range(points)])
            else:
                data_connector.cb_append_data_array([candle(x + xi) for xi in range(points)],
                                                    list(range(x, x + points)))
        x += 1
        sleep(0.01)


def category_generator(*data_connectors, categories: List, flip: bool = False):
    """Category generator"""
    x = 0
    while running:
        x += 1
        for data_connector in data_connectors:
            random_categories = random.sample(categories, random.randint(0, len(categories)))
            if flip:
                data_connector.cb_append_data_point(x * 0.01, random_categories)
            else:
                data_connector.cb_append_data_point(random_categories, x * 0.01)
        sleep(0.01)


def category_array_generator(*data_connectors, categories: List, flip: bool = False, points=10):
    """Category generator"""
    x = 0
    while running:
        for data_connector in data_connectors:
            random_categories = [random.sample(categories, random.randint(0, len(categories))) for xi in range(points)]
            if flip:
                data_connector.cb_append_data_array(list(range(x, x + points)), random_categories)
            else:
                data_connector.cb_append_data_array(random_categories, list(range(x, x + points)))
        x += 1
        sleep(0.01)


def colors():
    """Primitive color cycler"""
    while True:
        for r in range(50, 250, 2):
            for g in range(50, 250, 2):
                for b in range(50, 250, 2):
                    yield QColor(r, g, b)
                for c in range(50, 250, 2):
                    yield QColor(r, g, b - c)


# Initiate color generator
colors = colors()
# Connect SIGINT with stop function
signal.signal(signal.SIGINT, lambda sig, frame: stop())



import signal
from threading import Thread

import pglive.examples_pyside6 as examples
from pglive.kwargs import Axis
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_axis import LiveAxis
from pglive.sources.live_categorized_bar_plot import LiveCategorizedBarPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Categorized Bar plot is displayed in this example.
"""
categories = ["On", "Off", "Idle", "Warning", "Failure"]
plot = LiveCategorizedBarPlot(categories,
                              category_color={"Failure": "r", "Warning": "orange", "Off": "silver", "Idle": "blue"})
# Make sure to give plot.categories to Axis.CATEGORIES param (it's dynamic list)
# If you're using static categories and your data won't yield not listed category you can use static categories list
left_axis = LiveAxis("left", **{Axis.TICK_FORMAT: Axis.CATEGORY, Axis.CATEGORIES: plot.categories})
bottom_axis = LiveAxis("bottom", **{Axis.TICK_FORMAT: Axis.DURATION})
win = LivePlotWidget(title="Categorized Bar Plot @ 5Hz", axisItems={'bottom': bottom_axis, 'left': left_axis})
win.addItem(plot)
data_connector = DataConnector(plot, max_points=50, update_rate=5)
win.show()

Thread(target=examples.category_generator, args=(data_connector,), kwargs={"categories": categories}).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



import pglive.examples_pyside6 as examples
import signal
from threading import Thread

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveVBarPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Vertical Bar plot is displayed in this example.
"""
win = LivePlotWidget(title="Vertical Bar Plot @ 100Hz")
plot = LiveVBarPlot(bar_width=1, brush="blue", pen="blue")
win.addItem(plot)

data_connector = DataConnector(plot, max_points=600)

win.show()

Thread(target=examples.sin_wave_generator, args=(data_connector,)).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



import signal
from threading import Thread

import pglive.examples_pyqt6 as examples
import pyqtgraph as pg  # type: ignore

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_axis_range import LiveAxisRange
from pglive.sources.live_plot import LiveLinePlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
We are plotting two signals in one plot with different plot_rate in this example.
Since LiveAxisRange is calculating range from plot_rate, it might result in unwanted results.
You can use ignore_auto_range flag for DataConnector.
If it's set to True, this DataConnector is not causing change of range.
You can in fact use it to all LiveAxisRanges and implement your own custom Range calculation.
"""
layout = pg.LayoutWidget()
layout.layout.setSpacing(0)

'''
We want to display two signals in different frequencies.
Pglive is calculating view in respect to all plots. This might lead to unwanted results with multiple plots in one view.
So to makes things looks nice, we use only 100Hz plot for view calculation and ignore slow 1Hz plot.
Like that we can even save some resources and speed up overall plotting performance. 
'''
widget = LivePlotWidget(title="Two signals plotting at different rates.",
                        x_range_controller=LiveAxisRange(roll_on_tick=100, offset_left=30),
                        y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
widget.x_range_controller.crop_left_offset_to_data = True
plot = LiveLinePlot(pen="red")
widget.addItem(plot)
plot2 = LiveLinePlot(pen="yellow")
widget.addItem(plot2)
layout.addWidget(widget, row=0, col=0)
data_connector = DataConnector(plot, max_points=6000, plot_rate=100)
data_connector2 = DataConnector(plot2, max_points=6000, plot_rate=1, ignore_auto_range=True)
layout.show()

Thread(target=examples.sin_wave_generator, args=(data_connector,)).start()
Thread(target=examples.cos_wave_generator, args=(data_connector2,)).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



import pglive.examples_pyside6 as examples
from math import ceil
from threading import Thread

import pyqtgraph as pg  # type: ignore

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveHBarPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Different update rate is demonstrated in this example.
Display four plots, each slower by 1/4 of previous update rate.
Update rate is set in Hz unit.
"""
layout = pg.LayoutWidget()
args = []
# Initial rate of 100Hz
update_rate = 100.
max_len = 600.
# Initial delta Y is 1
bar_height = 1
for index in range(4):
    widget = LivePlotWidget(title=f"Horizontal Bar Plot @ {update_rate}Hz")
    plot = LiveHBarPlot(bar_height=bar_height, brush="green", pen="green")
    widget.addItem(plot)
    layout.addWidget(widget)
    args.append(DataConnector(plot, max_points=ceil(max_len), update_rate=update_rate))
    # divide all important parameters by 4
    update_rate /= 4.
    max_len /= 4.
    # bar height depends on Y distance, that's why we should multiply it by 4
    # if we leave it at 1, we get smaller bars
    bar_height *= 4

layout.show()

Thread(target=examples.sin_wave_generator, args=args).start()
examples.app.exec()
examples.stop()



import pglive.examples_pyside6 as examples
import signal
from threading import Thread

import pyqtgraph as pg  # type: ignore
from pglive.kwargs import Axis
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_axis import LiveAxis
from pglive.sources.live_axis_range import LiveAxisRange
from pglive.sources.live_plot import LiveLinePlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
Live plot range is used to increase plotting performance in this example.
"""

layout = pg.LayoutWidget()
layout.layout.setSpacing(0)
args = []
args2 = []

'''
Move view to the right on every 300 ticks (data update).
Y range is automatically adjudicating every tick.
'''
widget = LivePlotWidget(title=f"Rolling X view range, auto Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(roll_on_tick=300))
plot = LiveLinePlot(pen="green")
widget.addItem(plot)
layout.addWidget(widget)
args.append(DataConnector(plot, max_points=300, update_rate=100))

'''
Move view to the right on every 300 ticks (data update).
Y range is fixed to -1 and 1.
'''

widget = LivePlotWidget(title=f"Rolling X view range, fixed Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(roll_on_tick=300),
                        y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
plot = LiveLinePlot(pen="green")
widget.addItem(plot)
layout.addWidget(widget)
args.append(DataConnector(plot, max_points=300, update_rate=100))

'''
Move view to the right on every 150 ticks (data update).
Y range is fixed to -1 and 1.
In this case we have max. of 300 points in the buffer, but view is fixed to the last 150.
'''

widget = LivePlotWidget(title=f"Rolling X at half of max points, fixed Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(roll_on_tick=150),
                        y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
plot = LiveLinePlot(pen="green")
widget.addItem(plot)
layout.addWidget(widget, row=1, col=0)
args.append(DataConnector(plot, max_points=300, update_rate=100))

'''
Move view to the right on every 150 ticks (data update).
Y range is fixed to -1 and 1.
In this case we set offset_left to 0.5 (50% of last 150 points in the DataConnector's buffer).
Like this we can always display fixed first 150 ticks on every roll (not starting with blank view).
Of course we need to set max_points to value 2 * roll_on_ticks.
'''

widget = LivePlotWidget(title=f"Rolling X at half of max points with 50% left offset, fixed Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(roll_on_tick=150, offset_left=0.5),
                        y_range_controller=LiveAxisRange(fixed_range=[-1, 1]))
plot = LiveLinePlot(pen="green")
widget.addItem(plot)
layout.addWidget(widget, row=1, col=1)
args.append(DataConnector(plot, max_points=300, update_rate=100))

'''
Move view to the top on every 300 ticks (data update).
X range is fixed to -1 and 1.
'''

widget = LivePlotWidget(title=f"Rolling X at half of max points with 50% left offset, fixed Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(fixed_range=[-1, 1]),
                        y_range_controller=LiveAxisRange(roll_on_tick=300))
plot = LiveLinePlot(pen="white")
widget.addItem(plot)
layout.addWidget(widget, row=0, col=2)
args2.append(DataConnector(plot, max_points=300, update_rate=100))

'''
Move view to the top on every 150 ticks (data update).
X range is fixed to -1 and 1.
'''

widget = LivePlotWidget(title=f"Rolling X at half of max points with 50% left offset, fixed Y range @ 100Hz",
                        x_range_controller=LiveAxisRange(fixed_range=[-1, 1]),
                        y_range_controller=LiveAxisRange(roll_on_tick=150, offset_left=0.5))
plot = LiveLinePlot(pen="white")
widget.addItem(plot)
layout.addWidget(widget, row=1, col=2)
args2.append(DataConnector(plot, max_points=300, update_rate=100))

"""
Move view to the right on every 600 ticks (data update).
Y range is automatically adjudicating every tick.
"""
left_axis = LiveAxis("left", axisPen="red", textPen="red")
bottom_axis = LiveAxis("bottom", axisPen="green", textPen="green", **{Axis.TICK_FORMAT: Axis.TIME})
time_axis_plot_widget = LivePlotWidget(title="Rolling X view range, auto Y range @ 100Hz",
                                       axisItems={'bottom': bottom_axis, 'left': left_axis},
                                       x_range_controller=LiveAxisRange(roll_on_tick=600))
plot = LiveLinePlot(pen="yellow")
time_axis_plot_widget.addItem(plot)
args.append(DataConnector(plot, max_points=600))
layout.addWidget(time_axis_plot_widget, row=2, col=0)

"""
Move view to the right on every 300 ticks (data update).
Y range is automatically adjudicating every tick.
In this case we set offset_left to 0.5 (50% of last 300 points in the DataConnector's buffer).
Like this we can always display fixed first 300 ticks on every roll (not starting with blank view).
Of course we need to set max_points to value 2 * roll_on_ticks.
"""
left_axis = LiveAxis("left", axisPen="red", textPen="red")
bottom_axis = LiveAxis("bottom", axisPen="green", textPen="green", **{Axis.TICK_FORMAT: Axis.TIME})
time_axis_plot_widget = LivePlotWidget(
    title="Rolling X at half of max points with 50% left offset, fixed Y range @ 100Hz",
    axisItems={'bottom': bottom_axis, 'left': left_axis},
    x_range_controller=LiveAxisRange(roll_on_tick=300, offset_left=0.5))
plot = LiveLinePlot(pen="yellow")
time_axis_plot_widget.addItem(plot)
args.append(DataConnector(plot, max_points=600))
layout.addWidget(time_axis_plot_widget, row=2, col=1)

"""
Move view to the right on every 100 ticks (data update).
Y range is automatically adjudicating every tick.
"""
left_axis = LiveAxis("left", axisPen="red", textPen="red")
bottom_axis = LiveAxis("bottom", axisPen="green", textPen="green", **{Axis.TICK_FORMAT: Axis.TIME})
time_axis_plot_widget = LivePlotWidget(title="Rolling X at 1/6 of max points, fixed Y range @ 100Hz",
                                       axisItems={'bottom': bottom_axis, 'left': left_axis},
                                       x_range_controller=LiveAxisRange(roll_on_tick=100))
plot = LiveLinePlot(pen="yellow")
time_axis_plot_widget.addItem(plot)
args.append(DataConnector(plot, max_points=600))
layout.addWidget(time_axis_plot_widget, row=2, col=2)

# ---

layout.show()
Thread(target=examples.sin_wave_generator, args=args).start()
Thread(target=examples.sin_wave_generator, args=args2, kwargs={"flip": True}).start()

signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
examples.app.exec()
examples.stop()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'win_template.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(579, 330)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.label = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.plot_widget = LivePlotWidget(self.centralwidget)
        self.plot_widget.setObjectName("plot_widget")
        self.gridLayout.addWidget(self.plot_widget, 1, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 579, 25))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label.setText(_translate("MainWindow", "PyQt5 Designer example"))
from pglive.sources.live_plot_widget import LivePlotWidget



import datetime
from math import floor
from typing import Any, Optional, List

import pyqtgraph as pg  # type: ignore
from pyqtgraph import debug as debug, mkPen, getConfigOption  # type: ignore
from pyqtgraph.Qt import QtGui  # type: ignore

from pglive.kwargs import Axis
from pglive.sources.utils import get_scaled_time_duration


class LiveAxis(pg.AxisItem):
    """Implements live axis"""

    def __init__(self, orientation, pen=None, textPen=None, axisPen=None, linkView=None, parent=None, maxTickLength=-5,
                 showValues=True, text='', units='', unitPrefix='', **kwargs: Any) -> None:
        self.tick_position_indexes: Optional[List] = None
        super().__init__(orientation, pen=pen, textPen=textPen, linkView=linkView, parent=parent,
                         maxTickLength=maxTickLength, showValues=showValues, text=text, units=units,
                         unitPrefix=unitPrefix, **kwargs)
        # Fixing pyqtgraph bug, not setting textPen properly
        if textPen is None:
            self.setTextPen()
        else:
            self.setTextPen(textPen)
        # Set axisPen
        if axisPen is None:
            self.setAxisPen()
        else:
            self.setAxisPen(axisPen)
        # Tick format
        self.tick_format = kwargs.get(Axis.TICK_FORMAT, None)
        self.categories = kwargs.get(Axis.CATEGORIES, [])
        self.df_short = kwargs.get(Axis.DURATION_FORMAT, Axis.DF_SHORT) == Axis.DF_SHORT
        if self.tick_format == Axis.CATEGORY:
            # Override ticks spacing and set spacing 1 with step 1
            self.setTickSpacing(1, 1)

    def axisPen(self) -> QtGui.QPen:
        """Get axis pen"""
        if self._axisPen is None:
            return mkPen(getConfigOption('foreground'))
        return mkPen(self._axisPen)

    def setAxisPen(self, *args: Any, **kwargs: Any) -> None:
        """
        Set axis pen used for drawing axis line.
        If no arguments are given, the default foreground color will be used.
        """
        self.picture = None
        if args or kwargs:
            self._axisPen = mkPen(*args, **kwargs)
        else:
            self._axisPen = mkPen(getConfigOption('foreground'))
        self._updateLabel()

    def tickStrings(self, values: List, scale: float, spacing: float) -> List:
        """Convert ticks into final strings"""
        if self.tick_position_indexes:
            try:
                values = [self.tick_position_indexes[int(value - 1)] for value in values]
            except IndexError:
                pass
        if self.tick_format == Axis.DATETIME:
            # Convert tick to Datetime
            tick_strings = [datetime.datetime.fromtimestamp(value).strftime("%Y-%m-%d %H:%M:%S") for value in values]
        elif self.tick_format == Axis.TIME:
            # Convert tick to Time
            tick_strings = [datetime.datetime.fromtimestamp(value).strftime("%H:%M:%S") for value in values]
        elif self.tick_format == Axis.DURATION:
            # Convert tick to Time duration
            tick_strings = [get_scaled_time_duration(value, short=self.df_short) for value in values]
        elif self.tick_format == Axis.CATEGORY:
            # Convert tick to Category name
            tick_strings = []
            for value in values:
                try:
                    value += 0.5
                    tick_strings.append(self.categories[int(value)] if floor(value) >= 0 else "")
                except IndexError:
                    tick_strings.append("")
        else:
            # No specific format
            tick_strings = super().tickStrings(values, scale, spacing)
        return tick_strings

    def drawPicture(self, p, axisSpec, tickSpecs, textSpecs) -> None:
        profiler = debug.Profiler()

        p.setRenderHint(p.RenderHint.Antialiasing, False)
        p.setRenderHint(p.RenderHint.TextAntialiasing, True)

        # draw long line along axis
        pen, p1, p2 = axisSpec
        # Use axis pen to draw axis line
        p.setPen(self.axisPen())
        p.drawLine(p1, p2)
        # Switch back to normal pen
        p.setPen(pen)
        # p.translate(0.5,0)  ## resolves some damn pixel ambiguity

        # draw ticks
        for pen, p1, p2 in tickSpecs:
            p.setPen(pen)
            p.drawLine(p1, p2)
        profiler('draw ticks')

        # Draw all text
        if self.style['tickFont'] is not None:
            p.setFont(self.style['tickFont'])
        p.setPen(self.textPen())
        bounding = self.boundingRect().toAlignedRect()
        p.setClipRect(bounding)
        for rect, flags, text in textSpecs:
            p.drawText(rect, int(flags), text)

        profiler('draw text')



from typing import List, Tuple

import pglive.examples_pyqt5 as examples
from threading import Thread

import pyqtgraph as pg  # type: ignore

from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveHBarPlot, LiveVBarPlot, LiveLinePlot, LiveScatterPlot
from pglive.sources.live_plot_widget import LivePlotWidget

"""
All basic plot types are displayed in this example.
"""

layout = pg.LayoutWidget()
args = []
plots: List[Tuple[str, LivePlotWidget]] = [
    ("Horizontal Bar Plot", LiveHBarPlot(bar_height=2, brush="blue", pen="blue")),
    ("Vertical Bar Plot", LiveVBarPlot(bar_width=2, brush="green", pen="green")),
    ("Line Plot", LiveLinePlot(pen="red")),
    ("Scatter Plot", LiveScatterPlot(brush="yellow", pen="yellow")),
]

for title, plot in plots:
    widget = LivePlotWidget(title=f"{title} @ 50Hz")
    widget.addItem(plot)
    layout.addWidget(widget)
    args.append(DataConnector(plot, max_points=300, update_rate=50))

layout.show()

Thread(target=examples.sin_wave_generator, args=args).start()
examples.app.exec()
examples.stop()



from __future__ import annotations

from typing import Union, Optional, Any, List, Dict, TYPE_CHECKING

import pyqtgraph as pg  # type: ignore
from pyqtgraph import ViewBox  # type: ignore
from pyqtgraph.Qt import QtCore  # type: ignore
from pyqtgraph.Qt import QtGui  # type: ignore

from pglive.kwargs import Crosshair
from pglive.sources.live_axis import LiveAxis
from pglive.sources.live_axis_range import LiveAxisRange

if TYPE_CHECKING:
    from pglive.sources.data_connector import DataConnector


class LivePlotWidget(pg.PlotWidget):
    """Implements main plot widget for all live plots"""
    mouse_position: Optional[QtCore.QPointF] = None
    sig_crosshair_moved = QtCore.Signal(QtCore.QPointF)
    sig_crosshair_out = QtCore.Signal()
    sig_crosshair_in = QtCore.Signal()

    def __init__(self, parent=None, background: str = 'default', plotItem=None,
                 x_range_controller: Optional[LiveAxisRange] = None,
                 y_range_controller: Optional[LiveAxisRange] = None, **kwargs: Any) -> None:
        # Make sure we have LiveAxis in the bottom
        if "axisItems" not in kwargs:
            kwargs["axisItems"] = {"bottom": LiveAxis("bottom")}
        elif "bottom" not in kwargs["axisItems"]:
            kwargs["axisItems"]["bottom"] = LiveAxis("bottom")
        assert isinstance(kwargs["axisItems"]["bottom"], LiveAxis)
        self.x_range_controller = LiveAxisRange() if x_range_controller is None else x_range_controller
        self.y_range_controller = LiveAxisRange() if y_range_controller is None else y_range_controller
        self.manual_range = False

        super().__init__(parent=parent, background=background, plotItem=plotItem, **kwargs)
        self.final_x_range: List[float] = [self.viewRect().x(), self.viewRect().width()]
        self.final_y_range: List[float] = [self.viewRect().y(), self.viewRect().height()]
        self.life_ranges: Dict = {}

        self.crosshair_enabled = kwargs.get(Crosshair.ENABLED, False)
        self.crosshair_items: List = []
        self.crosshair_x_axis = kwargs.get(Crosshair.X_AXIS, "bottom")
        self.crosshair_y_axis = kwargs.get(Crosshair.Y_AXIS, "left")
        if self.crosshair_enabled:
            self.add_crosshair(kwargs.get(Crosshair.LINE_PEN, None),
                               kwargs.get(Crosshair.TEXT_KWARGS, {}))
        self.getPlotItem().autoBtn.clicked.disconnect()
        self.getPlotItem().autoBtn.clicked.connect(self.auto_btn_clicked)

        # Override addItem method
        def addItem(*args: Any, **kwargs: Any) -> None:
            if hasattr(args[0], "_vl_kwargs") and args[0]._vl_kwargs is not None:
                self.plotItem.addItem(args[0]._vl_kwargs["line"], ignoreBounds=True)
                self.plotItem.addItem(args[0]._vl_kwargs["text"], ignoreBounds=True)
            if hasattr(args[0], "_hl_kwargs") and args[0]._hl_kwargs is not None:
                self.plotItem.addItem(args[0]._hl_kwargs["line"], ignoreBounds=True)
                self.plotItem.addItem(args[0]._hl_kwargs["text"], ignoreBounds=True)
            if hasattr(args[0], "update_leading_line"):
                setattr(args[0], "x_format", self.x_format)
                setattr(args[0], "y_format", self.y_format)
            self.plotItem.addItem(*args)
            args[0].plot_widget = self

        def removeItem(*args):
            for item in args:
                if hasattr(item, "data_connector"):
                    self.x_range_controller.remove_data_connector(item.data_connector)
                    self.y_range_controller.remove_data_connector(item.data_connector)
            return self.plotItem.removeItem(*args)

        self.disableAutoRange()
        self.getPlotItem().vb.setRange = self.set_range
        self.getPlotItem().vb.sigRangeChangedManually.connect(self.sm)
        self.addItem = addItem
        self.removeItem = removeItem

    def sm(self, *args, **kwargs) -> None:
        self.manual_range = True

    def _update_crosshair_position(self) -> None:
        """Update position of crosshair based on mouse position"""
        if self.mouse_position:
            mouse_point = self.plotItem.vb.mapSceneToView(self.mouse_position)
            # Move crosshair to mouse pointer position
            self.vLine.setPos(mouse_point.x())
            self.hLine.setPos(mouse_point.y())

            self.x_value_label.setText(f"X = {self.x_format(mouse_point.x())}")
            self.y_value_label.setText(f"Y = {self.y_format(mouse_point.y())}")

            x_pos = self.mouse_position.x() + 5
            y_pos = self.mouse_position.y() - self.x_value_label.boundingRect().height()

            # Resolve position of crosshair text to be in view of the plot
            text_height = self.x_value_label.boundingRect().height()
            if self.mouse_position.x() + self.x_value_label.boundingRect().width() > self.plotItem.width():
                x_pos = self.mouse_position.x() - 5 - self.x_value_label.boundingRect().width()
            elif self.mouse_position.x() + self.y_value_label.boundingRect().width() > self.plotItem.width():
                x_pos = self.mouse_position.x() - 5 - self.x_value_label.boundingRect().width()
            if self.mouse_position.y() - self.x_value_label.boundingRect().height() - 2 * text_height - 5 < 0:
                y_pos += text_height * 2 + 5

            # Move X marker to position
            x_marker_point = self.plotItem.vb.mapSceneToView(
                QtCore.QPointF(x_pos, y_pos - self.x_value_label.boundingRect().height()))
            self.x_value_label.setPos(x_marker_point.x(), x_marker_point.y())

            # Move Y marker to position
            y_marker_point = self.plotItem.vb.mapSceneToView(QtCore.QPointF(x_pos, y_pos))
            self.y_value_label.setPos(y_marker_point.x(), y_marker_point.y())

            # Emit crosshair moved signal
            self.sig_crosshair_moved.emit(mouse_point)

    def add_crosshair(self, crosshair_pen: QtGui.QPen, crosshair_text_kwargs: dict) -> None:
        """Add crosshair into plot"""
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=crosshair_pen)
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=crosshair_pen)
        self.x_value_label = pg.TextItem(**crosshair_text_kwargs)
        self.y_value_label = pg.TextItem(**crosshair_text_kwargs)
        # All Crosshair items
        self.crosshair_items = [self.hLine, self.vLine, self.x_value_label, self.y_value_label]
        for item in self.crosshair_items:
            # Make sure, that every crosshair item is painted on top of everything
            item.setZValue(999)
            self.addItem(item, ignoreBounds=True)

        self.crosshair_enabled = True
        # Hide crosshair at the beginning
        self.hide_crosshair()

    def remove_crosshair(self):
        """Remove crosshair from the plot"""
        for item in self.crosshair_items:
            self.removeItem(item)
        self.crosshair_items = []
        self.crosshair_enabled = False

    def x_format(self, value: Union[int, float]) -> str:
        """X tick format"""
        try:
            # Get crosshair X str format from bottom tick axis format
            return self.getPlotItem().axes[self.crosshair_x_axis]["item"].tickStrings((value,), 0, 1)[0]
        except Exception:
            return str(round(value, 4))

    def y_format(self, value: Union[int, float]) -> str:
        """Y tick format"""
        try:
            # Get crosshair Y str format from left tick axis format
            return self.getPlotItem().axes[self.crosshair_y_axis]["item"].tickStrings((value,), 0, 1)[0]
        except Exception:
            return str(round(value, 4))

    def leaveEvent(self, ev: QtCore.QEvent) -> None:
        """Mouse left PlotWidget"""
        if self.crosshair_enabled:
            self.hide_crosshair()
        self.mouse_position = None
        self.sig_crosshair_out.emit()
        super().leaveEvent(ev)

    def enterEvent(self, ev: QtCore.QEvent) -> None:
        """Mouse enter PlotWidget"""
        if self.crosshair_enabled:
            self.show_crosshair()
        self.sig_crosshair_in.emit()
        super().enterEvent(ev)

    def mouseMoveEvent(self, ev: QtCore.QEvent) -> None:
        """Mouse moved in PlotWidget"""
        if pg.Qt.QT_LIB == pg.Qt.PYQT6:
            ev_pos = ev.position()
        else:
            ev_pos = ev.pos()

        if self.crosshair_enabled and self.sceneBoundingRect().contains(ev_pos):
            self.mouse_position = ev_pos
            self._update_crosshair_position()
        super().mouseMoveEvent(ev)

    def paintEvent(self, ev: QtCore.QEvent) -> None:
        """Update crosshair position when replot"""
        self._update_crosshair_position()
        return super().paintEvent(ev)

    def hide_crosshair(self) -> None:
        """Hide crosshair items"""
        for item in self.crosshair_items:
            item.hide()

    def show_crosshair(self) -> None:
        """Show crosshair items"""
        for item in self.crosshair_items:
            item.show()

    def auto_btn_clicked(self) -> None:
        """Controls auto button"""
        self.manual_range = False
        self.set_range(xRange=self.final_x_range, yRange=self.final_y_range)

    def slot_roll_tick(self, data_connector, tick: int) -> None:
        if data_connector.ignore_auto_range:
            # Don't calculate range for this DataConnector
            return
        elif not data_connector.plot.isVisible():
            # Calculate range for this DataConnector, but don't display plot
            self.x_range_controller.get_x_range(data_connector, tick)
            self.y_range_controller.get_y_range(data_connector, tick)
            return

        final_x_range = self.x_range_controller.get_x_range(data_connector, tick)
        final_y_range = self.y_range_controller.get_y_range(data_connector, tick)

        if self.final_x_range != final_x_range or self.final_y_range != final_y_range:
            self.final_x_range = final_x_range
            self.final_y_range = final_y_range
            if not self.manual_range:
                self.set_range(xRange=final_x_range, yRange=final_y_range)

    def slot_connector_toggle(self, data_connector: DataConnector, flag: bool) -> None:
        """Reset both range controllers when data_connector sets new data"""
        self.x_range_controller.ignore_connector(data_connector, flag)
        self.y_range_controller.ignore_connector(data_connector, flag)

        """
        After connector toggle, we must update view again.
        Calculate new x and y range.
        """
        final_x_range = self.x_range_controller.recalculate_x_range()
        final_y_range = self.y_range_controller.recalculate_y_range()

        # Force to update view if x or y range changed
        if self.final_x_range != final_x_range or self.final_y_range != final_y_range:
            self.final_x_range = final_x_range
            self.final_y_range = final_y_range
            if not self.manual_range:
                self.set_range(xRange=final_x_range, yRange=final_y_range)

    def set_range(self, *args, **kwargs):
        kwargs["disableAutoRange"] = True
        ViewBox.setRange(self.getPlotItem().vb, *args, **kwargs)



import signal
from threading import Thread

from PyQt5.QtWidgets import QMainWindow

import pglive.examples_pyqt5 as examples
from pglive.examples_pyqt5.designer_example.win_template import Ui_MainWindow
from pglive.sources.data_connector import DataConnector
from pglive.sources.live_plot import LiveLinePlot


class MainWindow(QMainWindow, Ui_MainWindow):
    """Create main window from template"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)


win = MainWindow()
plot = LiveLinePlot()
data_connector = DataConnector(plot, max_points=600)
win.plot_widget.addItem(plot)

Thread(target=examples.sin_wave_generator, args=(data_connector,)).start()
signal.signal(signal.SIGINT, lambda sig, frame: examples.stop())
win.show()
examples.app.exec()
examples.stop()
